{
    "quota_class_set": {
        "backup_gigabytes": 1000,
        "backups": 10,
        "gigabytes": 1000,
        "gigabytes___DEFAULT__": -1,
        "groups": 10,
        "id": "test_class",
        "per_volume_gigabytes": -1,
        "snapshots": 10,
        "snapshots___DEFAULT__": -1,
        "volumes": 10,
        "volumes___DEFAULT__": -1
    }
}