{
    "extensions": [
        {
            "alias": "os-hosts",
            "description": "Admin-only host administration.",
            "links": [],
            "name": "Hosts",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-vol-tenant-attr",
            "description": "Expose the internal project_id as an attribute of a volume.",
            "links": [],
            "name": "VolumeTenantAttribute",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-quota-sets",
            "description": "Quota management support.",
            "links": [],
            "name": "Quotas",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-availability-zone",
            "description": "Describe Availability Zones.",
            "links": [],
            "name": "AvailabilityZones",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-encryption-metadata",
            "description": "Volume encryption metadata retrieval support.",
            "links": [],
            "name": "VolumeEncryptionMetadata",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "backups",
            "description": "Backups support.",
            "links": [],
            "name": "Backups",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-snapshot-actions",
            "description": "Enable snapshot manager actions.",
            "links": [],
            "name": "SnapshotActions",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-actions",
            "description": "Enable volume actions.",
            "links": [],
            "name": "VolumeActions",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-snapshot-manage",
            "description": "Allows existing backend storage to be 'managed' by Cinder.",
            "links": [],
            "name": "SnapshotManage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-unmanage",
            "description": "Enable volume unmanage operation.",
            "links": [],
            "name": "VolumeUnmanage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "consistencygroups",
            "description": "consistency groups support.",
            "links": [],
            "name": "Consistencygroups",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-vol-host-attr",
            "description": "Expose host as an attribute of a volume.",
            "links": [],
            "name": "VolumeHostAttribute",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "encryption",
            "description": "Encryption support for volume types.",
            "links": [],
            "name": "VolumeTypeEncryption",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-vol-image-meta",
            "description": "Show image metadata associated with the volume.",
            "links": [],
            "name": "VolumeImageMetadata",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-types-manage",
            "description": "Types manage support.",
            "links": [],
            "name": "TypesManage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "capabilities",
            "description": "Capabilities support.",
            "links": [],
            "name": "Capabilities",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "cgsnapshots",
            "description": "cgsnapshots support.",
            "links": [],
            "name": "Cgsnapshots",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-types-extra-specs",
            "description": "Type extra specs support.",
            "links": [],
            "name": "TypesExtraSpecs",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-used-limits",
            "description": "Provide data on limited resources that are being used.",
            "links": [],
            "name": "UsedLimits",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-vol-mig-status-attr",
            "description": "Expose migration_status as an attribute of a volume.",
            "links": [],
            "name": "VolumeMigStatusAttribute",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-type-access",
            "description": "Volume type access support.",
            "links": [],
            "name": "VolumeTypeAccess",
            "updated": "%(isotime)s"
        },
        {
            "alias": "os-extended-services",
            "description": "Extended services support.",
            "links": [],
            "name": "ExtendedServices",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-extended-snapshot-attributes",
            "description": "Extended SnapshotAttributes support.",
            "links": [],
            "name": "ExtendedSnapshotAttributes",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-snapshot-unmanage",
            "description": "Enable volume unmanage operation.",
            "links": [],
            "name": "SnapshotUnmanage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "qos-specs",
            "description": "QoS specs support.",
            "links": [],
            "name": "Qos_specs_manage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-quota-class-sets",
            "description": "Quota classes management support.",
            "links": [],
            "name": "QuotaClasses",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-transfer",
            "description": "Volume transfer management support.",
            "links": [],
            "name": "VolumeTransfer",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-volume-manage",
            "description": "Allows existing backend storage to be 'managed' by Cinder.",
            "links": [],
            "name": "VolumeManage",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-admin-actions",
            "description": "Enable admin actions.",
            "links": [],
            "name": "AdminActions",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "os-services",
            "description": "Services support.",
            "links": [],
            "name": "Services",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "scheduler-stats",
            "description": "Scheduler stats support.",
            "links": [],
            "name": "Scheduler_stats",
            "updated": "%(extension_update)s"
        },
        {
            "alias": "OS-SCH-HNT",
            "description": "Pass arbitrary key/value pairs to the scheduler.",
            "links": [],
            "name": "SchedulerHints",
            "updated": "%(extension_update)s"
        }
    ]
}