.. _cinder-netapp_cdot_nfs:

.. list-table:: Description of NetApp cDOT NFS driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``expiry_thres_minutes`` = ``720``
     - (Integer) This option specifies the threshold for last access time for images in the NFS image cache. When a cache cleaning cycle begins, images in the cache that have not been accessed in the last M minutes, where M is the value of this parameter, will be deleted from the cache to create free space on the NFS share.
   * - ``netapp_copyoffload_tool_path`` = ``None``
     - (String) This option specifies the path of the NetApp copy offload tool binary. Ensure that the binary has execute permissions set which allow the effective user of the cinder-volume process to execute the file.
   * - ``netapp_host_type`` = ``None``
     - (String) This option defines the type of operating system for all initiators that can access a LUN. This information is used when mapping LUNs to individual hosts or groups of hosts.
   * - ``netapp_login`` = ``None``
     - (String) Administrative user account name used to access the storage system or proxy server.
   * - ``netapp_lun_ostype`` = ``None``
     - (String) This option defines the type of operating system that will access a LUN exported from Data ONTAP; it is assigned to the LUN at the time it is created.
   * - ``netapp_password`` = ``None``
     - (String) Password for the administrative user account specified in the netapp_login option.
   * - ``netapp_pool_name_search_pattern`` = ``(.+)``
     - (String) This option is used to restrict provisioning to the specified pools. Specify the value of this option to be a regular expression which will be applied to the names of objects from the storage backend which represent pools in Cinder. This option is only utilized when the storage protocol is configured to use iSCSI or FC.
   * - ``netapp_replication_aggregate_map`` = ``None``
     - (Unknown) Multi opt of dictionaries to represent the aggregate mapping between source and destination back ends when using whole back end replication. For every source aggregate associated with a cinder pool (NetApp FlexVol), you would need to specify the destination aggregate on the replication target device. A replication target device is configured with the configuration option replication_device. Specify this option as many times as you have replication devices. Each entry takes the standard dict config form: netapp_replication_aggregate_map = backend_id:<name_of_replication_device_section>,src_aggr_name1:dest_aggr_name1,src_aggr_name2:dest_aggr_name2,...
   * - ``netapp_server_hostname`` = ``None``
     - (String) The hostname (or IP address) for the storage system or proxy server.
   * - ``netapp_server_port`` = ``None``
     - (Integer) The TCP port to use for communication with the storage system or proxy server. If not specified, Data ONTAP drivers will use 80 for HTTP and 443 for HTTPS.
   * - ``netapp_snapmirror_quiesce_timeout`` = ``3600``
     - (Integer) The maximum time in seconds to wait for existing SnapMirror transfers to complete before aborting during a failover.
   * - ``netapp_storage_family`` = ``ontap_cluster``
     - (String) The storage family type used on the storage system; the only valid value is ontap_cluster for using clustered Data ONTAP.
   * - ``netapp_storage_protocol`` = ``None``
     - (String) The storage protocol to be used on the data path with the storage system.
   * - ``netapp_transport_type`` = ``http``
     - (String) The transport protocol used when communicating with the storage system or proxy server.
   * - ``netapp_vserver`` = ``None``
     - (String) This option specifies the virtual storage server (Vserver) name on the storage cluster on which provisioning of block storage volumes should occur.
   * - ``thres_avl_size_perc_start`` = ``20``
     - (Integer) If the percentage of available space for an NFS share has dropped below the value specified by this option, the NFS image cache will be cleaned.
   * - ``thres_avl_size_perc_stop`` = ``60``
     - (Integer) When the percentage of available space on an NFS share has reached the percentage specified by this option, the driver will stop clearing files from the NFS image cache that have not been accessed in the last M minutes, where M is the value of the expiry_thres_minutes configuration option.
