{
    "volume_types": [
        {
            "description": "volume type 0002",
            "extra_specs": {
                "capabilities": "gpu"
            },
            "id": "%(uuid)s",
            "is_public": true,
            "name": "vol-type-002",
            "os-volume-type-access:is_public": true,
            "qos_specs_id": null
        },
        {
            "description": "volume type 0001",
            "extra_specs": {
                "capabilities": "gpu"
            },
            "id": "%(uuid)s",
            "is_public": true,
            "name": "vol-type-001",
            "os-volume-type-access:is_public": true,
            "qos_specs_id": null
        },
        {
            "description": "Default Volume Type",
            "extra_specs": {},
            "id": "%(uuid)s",
            "is_public": true,
            "name": "__DEFAULT__",
            "os-volume-type-access:is_public": true,
            "qos_specs_id": null
        }
    ]
}