{
    "snapshot": {
        "description": null,
        "metadata": null,
        "ref": {
            "source-name": "lvol0"
        },
        "name": null,
        "volume_id": "1df34919-aba7-4a1b-a614-3b409d71ac03"
    }
}